@@ -1,7 +1,7 @@
 """
     emitters.py
 
-    Copyright (c) 2013-2020 Snowplow Analytics Ltd. All rights reserved.
+    Copyright (c) 2013-2021 Snowplow Analytics Ltd. All rights reserved.
 
     This program is licensed to you under the Apache License Version 2.0,
     and you may not use this file except in compliance with the Apache License
@@ -15,41 +15,31 @@
     language governing permissions and limitations there under.
 
     Authors: Anuj More, Alex Dean, Fred Blundun, Paul Boocock
-    Copyright: Copyright (c) 2013-2020 Snowplow Analytics Ltd
+    Copyright: Copyright (c) 2013-2021 Snowplow Analytics Ltd
     License: Apache License Version 2.0
 """
 
-import json
+
 import logging
 import time
 import threading
-try:
-    # Python 2
-    from Queue import Queue
-except ImportError:
-    # Python 3
-    from queue import Queue
-
-from celery import Celery
-import redis
 import requests
-from contracts import contract, new_contract
+from typing import Optional, Union, Tuple
+from queue import Queue
 
 from snowplow_tracker.self_describing_json import SelfDescribingJson
+from snowplow_tracker.typing import PayloadDict, PayloadDictList, HttpProtocol, Method, SuccessCallback, FailureCallback
+from snowplow_tracker.contracts import one_of
 
+# logging
+logging.basicConfig()
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.INFO)
 
 DEFAULT_MAX_LENGTH = 10
 PAYLOAD_DATA_SCHEMA = "iglu:com.snowplowanalytics.snowplow/payload_data/jsonschema/1-0-4"
-
-new_contract("protocol", lambda x: x == "http" or x == "https")
-
-new_contract("method", lambda x: x == "get" or x == "post")
-
-new_contract("function", lambda x: hasattr(x, "__call__"))
-
-new_contract("redis", lambda x: isinstance(x, (redis.Redis, redis.StrictRedis)))
+PROTOCOLS = {"http", "https"}
+METHODS = {"get", "post"}
 
 
 class Emitter(object):
@@ -58,8 +48,17 @@ class Emitter(object):
         Supports both GET and POST requests
     """
 
-    @contract
-    def __init__(self, endpoint, protocol="http", port=None, method="get", buffer_size=None, on_success=None, on_failure=None, byte_limit=None):
+    def __init__(
+            self,
+            endpoint: str,
+            protocol: HttpProtocol = "http",
+            port: Optional[int] = None,
+            method: Method = "get",
+            buffer_size: Optional[int] = None,
+            on_success: Optional[SuccessCallback] = None,
+            on_failure: Optional[FailureCallback] = None,
+            byte_limit: Optional[int] = None,
+            request_timeout: Optional[Union[float, Tuple[float, float]]] = None) -> None:
         """
             :param endpoint:    The collector URL. Don't include "http://" - this is done automatically.
             :type  endpoint:    string
@@ -82,7 +81,14 @@ class Emitter(object):
             :type  on_failure:  function | None
             :param byte_limit:  The size event list after reaching which queued events will be flushed
             :type  byte_limit:  int | None
+            :param request_timeout: Timeout for the HTTP requests. Can be set either as single float value which
+                                     applies to both "connect" AND "read" timeout, or as tuple with two float values
+                                     which specify the "connect" and "read" timeouts separately
+            :type request_timeout:  float | tuple | None
         """
+        one_of(protocol, PROTOCOLS)
+        one_of(method, METHODS)
+
         self.endpoint = Emitter.as_collector_uri(endpoint, protocol, port, method)
 
         self.method = method
@@ -96,6 +102,7 @@ class Emitter(object):
         self.buffer = []
         self.byte_limit = byte_limit
         self.bytes_queued = None if byte_limit is None else 0
+        self.request_timeout = request_timeout
 
         self.on_success = on_success
         self.on_failure = on_failure
@@ -107,8 +114,11 @@ class Emitter(object):
         logger.info("Emitter initialized with endpoint " + self.endpoint)
 
     @staticmethod
-    @contract
-    def as_collector_uri(endpoint, protocol="http", port=None, method="get"):
+    def as_collector_uri(
+            endpoint: str,
+            protocol: HttpProtocol = "http",
+            port: Optional[int] = None,
+            method: Method = "get") -> str:
         """
             :param endpoint:  The raw endpoint provided by the user
             :type  endpoint:  string
@@ -120,6 +130,9 @@ class Emitter(object):
             :type  method:    method
             :rtype:           string
         """
+        if len(endpoint) < 1:
+            raise ValueError("No endpoint provided.")
+
         if method == "get":
             path = "/i"
         else:
@@ -129,8 +142,7 @@ class Emitter(object):
         else:
             return protocol + "://" + endpoint + ":" + str(port) + path
 
-    @contract
-    def input(self, payload):
+    def input(self, payload: PayloadDict) -> None:
         """
             Adds an event to the buffer.
             If the maximum size has been reached, flushes the buffer.
@@ -150,7 +162,7 @@ class Emitter(object):
             if self.reached_limit():
                 self.flush()
 
-    def reached_limit(self):
+    def reached_limit(self) -> bool:
         """
             Checks if event-size or bytes limit are reached
 
@@ -159,9 +171,9 @@ class Emitter(object):
         if self.byte_limit is None:
             return len(self.buffer) >= self.buffer_size
         else:
-            return self.bytes_queued >= self.byte_limit or len(self.buffer) >= self.buffer_size
+            return (self.bytes_queued or 0) >= self.byte_limit or len(self.buffer) >= self.buffer_size
 
-    def flush(self):
+    def flush(self) -> None:
         """
             Sends all events in the buffer to the collector.
         """
@@ -171,42 +183,55 @@ class Emitter(object):
             if self.bytes_queued is not None:
                 self.bytes_queued = 0
 
-    @contract
-    def http_post(self, data):
+    def http_post(self, data: str) -> bool:
         """
             :param data:  The array of JSONs to be sent
             :type  data:  string
         """
         logger.info("Sending POST request to %s..." % self.endpoint)
         logger.debug("Payload: %s" % data)
-        r = requests.post(self.endpoint, data=data, headers={'content-type': 'application/json; charset=utf-8'})
-        getattr(logger, "info" if self.is_good_status_code(r.status_code) else "warn")("POST request finished with status code: " + str(r.status_code))
-        return r
+        post_succeeded = False
+        try:
+            r = requests.post(
+                self.endpoint,
+                data=data,
+                headers={'Content-Type': 'application/json; charset=utf-8'},
+                timeout=self.request_timeout)
+            post_succeeded = Emitter.is_good_status_code(r.status_code)
+            getattr(logger, "info" if post_succeeded else "warning")("POST request finished with status code: " + str(r.status_code))
+        except requests.RequestException as e:
+            logger.warning(e)
 
-    @contract
-    def http_get(self, payload):
+        return post_succeeded
+
+    def http_get(self, payload: PayloadDict) -> bool:
         """
             :param payload:  The event properties
             :type  payload:  dict(string:*)
         """
         logger.info("Sending GET request to %s..." % self.endpoint)
         logger.debug("Payload: %s" % payload)
-        r = requests.get(self.endpoint, params=payload)
-        getattr(logger, "info" if self.is_good_status_code(r.status_code) else "warn")("GET request finished with status code: " + str(r.status_code))
-        return r
+        get_succeeded = False
+        try:
+            r = requests.get(self.endpoint, params=payload, timeout=self.request_timeout)
+            get_succeeded = Emitter.is_good_status_code(r.status_code)
+            getattr(logger, "info" if get_succeeded else "warning")("GET request finished with status code: " + str(r.status_code))
+        except requests.RequestException as e:
+            logger.warning(e)
 
-    def sync_flush(self):
+        return get_succeeded
+
+    def sync_flush(self) -> None:
         """
             Calls the flush method of the base Emitter class.
             This is guaranteed to be blocking, not asynchronous.
         """
         logger.debug("Starting synchronous flush...")
         Emitter.flush(self)
-        logger.info("Finished synchrous flush")
+        logger.info("Finished synchronous flush")
 
     @staticmethod
-    @contract
-    def is_good_status_code(status_code):
+    def is_good_status_code(status_code: int) -> bool:
         """
             :param status_code:  HTTP status code
             :type  status_code:  int
@@ -214,53 +239,43 @@ class Emitter(object):
         """
         return 200 <= status_code < 400
 
-    @contract
-    def send_events(self, evts):
+    def send_events(self, evts: PayloadDictList) -> None:
         """
             :param evts: Array of events to be sent
             :type  evts: list(dict(string:*))
         """
         if len(evts) > 0:
-            logger.info("Attempting to send %s requests" % len(evts))
+            logger.info("Attempting to send %s events" % len(evts))
+
             Emitter.attach_sent_timestamp(evts)
+            success_events = []
+            failure_events = []
+
             if self.method == 'post':
                 data = SelfDescribingJson(PAYLOAD_DATA_SCHEMA, evts).to_string()
-                post_succeeded = False
-                try:
-                    status_code = self.http_post(data).status_code
-                    post_succeeded = self.is_good_status_code(status_code)
-                except requests.RequestException as e:
-                    logger.warn(e)
-                if post_succeeded:
-                    if self.on_success is not None:
-                        self.on_success(len(evts))
-                elif self.on_failure is not None:
-                    self.on_failure(0, evts)
+                request_succeeded = self.http_post(data)
+                if request_succeeded:
+                    success_events += evts
+                else:
+                    failure_events += evts
 
             elif self.method == 'get':
-                success_count = 0
-                unsent_requests = []
                 for evt in evts:
-                    get_succeeded = False
-                    try:
-                        status_code = self.http_get(evt).status_code
-                        get_succeeded = self.is_good_status_code(status_code)
-                    except requests.RequestException as e:
-                        logger.warn(e)
-                    if get_succeeded:
-                        success_count += 1
+                    request_succeeded = self.http_get(evt)
+                    if request_succeeded:
+                        success_events += [evt]
                     else:
-                        unsent_requests.append(evt)
-                if len(unsent_requests) == 0:
-                    if self.on_success is not None:
-                        self.on_success(success_count)
-                elif self.on_failure is not None:
-                    self.on_failure(success_count, unsent_requests)
+                        failure_events += [evt]
+
+            if self.on_success is not None and len(success_events) > 0:
+                self.on_success(success_events)
+            if self.on_failure is not None and len(failure_events) > 0:
+                self.on_failure(len(success_events), failure_events)
+
         else:
             logger.info("Skipping flush since buffer is empty")
 
-    @contract
-    def set_flush_timer(self, timeout, flush_now=False):
+    def set_flush_timer(self, timeout: float, flush_now: bool = False) -> None:
         """
             Set an interval at which the buffer will be flushed
 
@@ -277,7 +292,7 @@ class Emitter(object):
         self.timer.daemon = True
         self.timer.start()
 
-    def cancel_flush_timer(self):
+    def cancel_flush_timer(self) -> None:
         """
             Abort automatic async flushing
         """
@@ -286,7 +301,7 @@ class Emitter(object):
             self.timer.cancel()
 
     @staticmethod
-    def attach_sent_timestamp(events):
+    def attach_sent_timestamp(events: PayloadDictList) -> None:
         """
             Attach (by mutating in-place) current timestamp in milliseconds
             as `stm` param
@@ -295,10 +310,11 @@ class Emitter(object):
             :type  events: list(dict(string:*))
             :rtype: None
         """
-        def update(e):
+        def update(e: PayloadDict) -> None:
             e.update({'stm': str(int(time.time()) * 1000)})
 
-        [update(event) for event in events]
+        for event in events:
+            update(event)
 
 
 class AsyncEmitter(Emitter):
@@ -306,18 +322,17 @@ class AsyncEmitter(Emitter):
         Uses threads to send HTTP requests asynchronously
     """
 
-    @contract
     def __init__(
-        self,
-        endpoint,
-        protocol="http",
-        port=None,
-        method="get",
-        buffer_size=None,
-        on_success=None,
-        on_failure=None,
-        thread_count=1,
-        byte_limit=None):
+            self,
+            endpoint: str,
+            protocol: HttpProtocol = "http",
+            port: Optional[int] = None,
+            method: Method = "get",
+            buffer_size: Optional[int] = None,
+            on_success: Optional[SuccessCallback] = None,
+            on_failure: Optional[FailureCallback] = None,
+            thread_count: int = 1,
+            byte_limit: Optional[int] = None) -> None:
         """
             :param endpoint:    The collector URL. Don't include "http://" - this is done automatically.
             :type  endpoint:    string
@@ -350,14 +365,14 @@ class AsyncEmitter(Emitter):
             t.daemon = True
             t.start()
 
-    def sync_flush(self):
+    def sync_flush(self) -> None:
         while True:
             self.flush()
             self.queue.join()
             if len(self.buffer) < 1:
                 break
 
-    def flush(self):
+    def flush(self) -> None:
         """
             Removes all dead threads, then creates a new thread which
             executes the flush method of the base Emitter class
@@ -368,75 +383,8 @@ class AsyncEmitter(Emitter):
             if self.bytes_queued is not None:
                 self.bytes_queued = 0
 
-    def consume(self):
+    def consume(self) -> None:
         while True:
             evts = self.queue.get()
             self.send_events(evts)
             self.queue.task_done()
-
-
-class CeleryEmitter(Emitter):
-    """
-        Uses a Celery worker to send HTTP requests asynchronously.
-        Works like the base Emitter class,
-        but on_success and on_failure callbacks cannot be set.
-    """
-    celery_app = None
-
-    def __init__(self, endpoint, protocol="http", port=None, method="get", buffer_size=None, byte_limit=None):
-        super(CeleryEmitter, self).__init__(endpoint, protocol, port, method, buffer_size, None, None, byte_limit)
-
-        try:
-            # Check whether a custom Celery configuration module named "snowplow_celery_config" exists
-            import snowplow_celery_config
-            self.celery_app = Celery()
-            self.celery_app.config_from_object(snowplow_celery_config)
-        except ImportError:
-            # Otherwise configure Celery with default settings
-            self.celery_app = Celery("Snowplow", broker="redis://guest@localhost//")
-
-        self.async_flush = self.celery_app.task(self.async_flush)
-
-    def flush(self):
-        """
-            Schedules a flush task
-        """
-        self.async_flush.delay()
-        logger.info("Scheduled a Celery task to flush the event queue")
-
-    def async_flush(self):
-        super(CeleryEmitter, self).flush()
-
-
-class RedisEmitter(object):
-    """
-        Sends Snowplow events to a Redis database
-    """
-    @contract
-    def __init__(self, rdb=None, key="snowplow"):
-        """
-            :param rdb:  Optional custom Redis database
-            :type  rdb:  redis | None
-            :param key:  The Redis key for the list of events
-            :type  key:  string
-        """
-        if rdb is None:
-            rdb = redis.StrictRedis()
-        self.rdb = rdb
-        self.key = key
-
-    @contract
-    def input(self, payload):
-        """
-            :param payload:  The event properties
-            :type  payload:  dict(string:*)
-        """
-        logger.debug("Pushing event to Redis queue...")
-        self.rdb.rpush(self.key, json.dumps(payload))
-        logger.info("Finished sending event to Redis.")
-
-    def flush(self):
-        logger.warn("The RedisEmitter class does not need to be flushed")
-
-    def sync_flush(self):
-        self.flush()
