@@ -1,7 +1,7 @@
 """
     payload.py
 
-    Copyright (c) 2013-2020 Snowplow Analytics Ltd. All rights reserved.
+    Copyright (c) 2013-2021 Snowplow Analytics Ltd. All rights reserved.
 
     This program is licensed to you under the Apache License Version 2.0,
     and you may not use this file except in compliance with the Apache License
@@ -15,20 +15,19 @@
     language governing permissions and limitations there under.
 
     Authors: Anuj More, Alex Dean, Fred Blundun, Paul Boocock
-    Copyright: Copyright (c) 2013-2020 Snowplow Analytics Ltd
+    Copyright: Copyright (c) 2013-2021 Snowplow Analytics Ltd
     License: Apache License Version 2.0
 """
 
-import random
-import time
 import json
 import base64
-from contracts import contract
+from typing import Any, Optional
+from snowplow_tracker.typing import PayloadDict, JsonEncoderFunction
 
 
 class Payload:
 
-    def __init__(self, dict_=None):
+    def __init__(self, dict_: Optional[PayloadDict] = None) -> None:
         """
             Constructor
         """
@@ -39,20 +38,18 @@ class Payload:
             for f in dict_:
                 self.nv_pairs[f] = dict_[f]
 
-
     """
     Methods to add to the payload
     """
 
-    def add(self, name, value):
+    def add(self, name: str, value: Any) -> None:
         """
             Add a name value pair to the Payload object
         """
         if not (value == "" or value is None):
             self.nv_pairs[name] = value
 
-    @contract
-    def add_dict(self, dict_, base64=False):
+    def add_dict(self, dict_: PayloadDict, base64: bool = False) -> None:
         """
             Add a dict of name value pairs to the Payload object
 
@@ -62,8 +59,13 @@ class Payload:
         for f in dict_:
             self.add(f, dict_[f])
 
-    @contract
-    def add_json(self, dict_, encode_base64, type_when_encoded, type_when_not_encoded):
+    def add_json(
+            self,
+            dict_: Optional[PayloadDict],
+            encode_base64: bool,
+            type_when_encoded: str,
+            type_when_not_encoded: str,
+            json_encoder: Optional[JsonEncoderFunction] = None) -> None:
         """
             Add an encoded or unencoded JSON to the payload
 
@@ -75,22 +77,24 @@ class Payload:
             :type   type_when_encoded:      string
             :param  type_when_not_encoded:  Name of the field when encode_base64 is not set
             :type   type_when_not_encoded:  string
+            :param json_encoder:            Custom JSON serializer that gets called on non-serializable object
+            :type  json_encoder:            function | None
         """
 
         if dict_ is not None and dict_ != {}:
 
-            json_dict = json.dumps(dict_, ensure_ascii=False)
+            json_dict = json.dumps(dict_, ensure_ascii=False, default=json_encoder)
 
             if encode_base64:
                 encoded_dict = base64.urlsafe_b64encode(json_dict.encode("utf-8"))
                 if not isinstance(encoded_dict, str):
                     encoded_dict = encoded_dict.decode("utf-8")
-                self.add(type_when_encoded, encoded_dict)                
+                self.add(type_when_encoded, encoded_dict)
 
             else:
                 self.add(type_when_not_encoded, json_dict)
 
-    def get(self):
+    def get(self) -> PayloadDict:
         """
             Returns the context dictionary from the Payload object
         """
